//
//	ID Engine
//	ID_IN.c - Input Manager
//	v1.0d1
//	By Jason Blochowiak
//

//
//	This module handles dealing with the various input devices
//
//	Depends on: Memory Mgr (for demo recording), Sound Mgr (for timing stuff),
//				User Mgr (for command line parms)
//
//	Globals:
//		LastScan - The keyboard scan code of the last key pressed
//		LastASCII - The ASCII value of the last key pressed
//	DEBUG - there are more globals
//

#include <memory.h>
#include "id_heads.h"
#include "system.h"
#include "input.h"
#include "console.h"

// joystick constants
#define	JoyScaleMax		32768
#define	JoyScaleShift	8
#define	MaxJoyValue		5000

//=============================================================================
//					GLOBAL VARIABLES
//=============================================================================

// configuration variables
wbool		MousePresent;
wbool		JoysPresent[MaxJoys];

// 	Global variables
wbool		Keyboard[NumCodes];
wbool		Paused;
char		LastASCII;
ScanCode	LastScan;
KeyboardDef	KbdDefs = {0x1d,0x38,0x47,0x48,0x49,0x4b,0x4d,0x4f,0x50,0x51};
JoystickDef	JoyDefs[MaxJoys];
ControlType	Controls[MaxPlayers];

Demo		DemoMode = demo_Off;
byte		*DemoBuffer;
word		DemoOffset,DemoSize;

//=============================================================================
//					LOCAL VARIABLES
//=============================================================================

byte	ASCIINames[] =
	// Unshifted ASCII for scan codes
{
//	 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
	0  ,27 ,'1','2','3','4','5','6','7','8','9','0','-','=',8  ,9  ,
	'q','w','e','r','t','y','u','i','o','p','[',']',13 ,0  ,'a','s',
	'd','f','g','h','j','k','l',';',39 ,'`',0  ,92 ,'z','x','c','v',
	'b','n','m',',','.','/',0  ,'*',0  ,' ',0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,'7','8','9','-','4','5','6','+','1',
	'2','3','0',127,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0	
},
				ShiftNames[] =
	// Shifted ASCII for scan codes
{
//	 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
	0  ,27 ,'!','@','#','$','%','^','&','*','(',')','_','+',8  ,9  ,
	'Q','W','E','R','T','Y','U','I','O','P','{','}',13 ,0  ,'A','S',
	'D','F','G','H','J','K','L',':',34 ,'~',0  ,'|','Z','X','C','V',
	'B','N','M','<','>','?',0  ,'*',0  ,' ',0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,'7','8','9','-','4','5','6','+','1',
	'2','3','0',127,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0   
},
				SpecialNames[] =
	// ASCII for 0xe0 prefixed codes
{
//	 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,13 ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,'/',0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0   
};

static	wbool		IN_Started;
static	wbool		CapsLock;
static	ScanCode	CurCode,LastCode;

static	Direction DirTable[] =
	// Quick lookup for total direction
{
	dir_NorthWest,	dir_North,	dir_NorthEast,
	dir_West,		dir_None,	dir_East,
	dir_SouthWest,	dir_South,	dir_SouthEast
};

static char *ParmStrings[] = {"nojoys","nomouse",nil};

//	Internal routines

void IN_KeyEvent(unsigned char key)
{
	static wbool special;
	byte c;

	if (Con_Key(key))
		return;

	if (key == 0xe0)
	// Special key prefix
		special = true;
	else if (key == 0xe1)
	// Handle Pause key
		Paused = true;
	else
	{
		if (key & 0x80)
	// Break code
		{
			key &= 0x7f;

			Keyboard[key] = false;
		}
		else
			// Make code
		{
			LastCode = CurCode;
			CurCode = LastScan = key;
			Keyboard[key] = true;

			if (special)
				c = SpecialNames[key];
			else
			{
				if (key == sc_CapsLock)
				{
					CapsLock ^= true;
					// DEBUG - make caps lock light work
				}

				if (Keyboard[sc_LShift] || Keyboard[sc_RShift])
	// If shifted
				{
					c = ShiftNames[key];
					if ((c >= 'A') && (c <= 'Z') && CapsLock)
						c += 'a' - 'A';
				}
				else
				{
					c = ASCIINames[key];
					if ((c >= 'a') && (c <= 'z') && CapsLock)
						c -= 'a' - 'A';
				}
			}
			if (c)
				LastASCII = c;
		}

		special = false;
	}
}

//	Gets the amount that the mouse has moved from the
//	mouse driver
static void INL_GetMouseDelta(int *x, int *y)
{
/*
	Mouse(MDelta);
	*x = _CX;
	*y = _DX;
*/
*x = *y = 0;
}

//	Gets the status of the mouse buttons from the mouse driver
static word INL_GetMouseButtons()
{
/*
	word buttons;

	Mouse(MButtons);
	buttons = _BX;
	return(buttons);
*/
return 0;
}

//	Reads the absolute position of the specified joystick
void IN_GetJoyAbs(word joy, word *xp, word *yp)
{
	S_PollJoyPosition(xp, yp);

/*	byte	xb,yb, xs,ys;
	word	x,y;

	x = y = 0;
	xs = joy? 2 : 0;		// Find shift value for x axis
	xb = 1 << xs;			// Use shift value to get x bit mask
	ys = joy? 3 : 1;		// Do the same for y axis
	yb = 1 << ys;

// Read the absolute joystick values
asm		pushf				// Save some registers
asm		push	si
asm		push	di
asm		cli					// Make sure an interrupt doesn't screw the timings


asm		mov		dx,0x201
asm		in		al,dx
asm		out		dx,al		// Clear the resistors

asm		mov		ah,[xb]		// Get masks into registers
asm		mov		ch,[yb]

asm		xor		si,si		// Clear count registers
asm		xor		di,di
asm		xor		bh,bh		// Clear high byte of bx for later

asm		push	bp			// Don't mess up stack frame
asm		mov		bp,MaxJoyValue

loop:
asm		in		al,dx		// Get bits indicating whether all are finished

asm		dec		bp			// Check bounding register
asm		jz		done		// We have a silly value - abort

asm		mov		bl,al		// Duplicate the bits
asm		and		bl,ah		// Mask off useless bits (in [xb])
asm		add		si,bx		// Possibly increment count register
asm		mov		cl,bl		// Save for testing later

asm		mov		bl,al
asm		and		bl,ch		// [yb]
asm		add		di,bx

asm		add		cl,bl
asm		jnz		loop 		// If both bits were 0, drop out

done:
asm     pop		bp

asm		mov		cl,[xs]		// Get the number of bits to shift
asm		shr		si,cl		//  and shift the count that many times

asm		mov		cl,[ys]
asm		shr		di,cl

asm		mov		[x],si		// Store the values into the variables
asm		mov		[y],di

asm		pop		di
asm		pop		si
asm		popf				// Restore the registers

	*xp = x;
	*yp = y;
*/
}

//	Returns the relative movement of the specified
//	joystick (from +/-127)
void INL_GetJoyDelta(word joy, int *dx, int *dy)
{
	word x, y;
	JoystickDef *def;

	IN_GetJoyAbs(joy, &x, &y);
	def = JoyDefs + joy;

	if (x < def->threshMinX)
	{
		if (x < def->joyMinX)
			x = def->joyMinX;

		x = -(x - def->threshMinX);
		x *= def->joyMultXL;
		x >>= JoyScaleShift;
		*dx = (x > 127)? -127 : -x;
	}
	else if (x > def->threshMaxX)
	{
		if (x > def->joyMaxX)
			x = def->joyMaxX;

		x = x - def->threshMaxX;
		x *= def->joyMultXH;
		x >>= JoyScaleShift;
		*dx = (x > 127)? 127 : x;
	}
	else
		*dx = 0;

	if (y < def->threshMinY)
	{
		if (y < def->joyMinY)
			y = def->joyMinY;

		y = -(y - def->threshMinY);
		y *= def->joyMultYL;
		y >>= JoyScaleShift;
		*dy = (y > 127)? -127 : -y;
	}
	else if (y > def->threshMaxY)
	{
		if (y > def->joyMaxY)
			y = def->joyMaxY;

		y = y - def->threshMaxY;
		y *= def->joyMultYH;
		y >>= JoyScaleShift;
		*dy = (y > 127)? 127 : y;
	}
	else
		*dy = 0;
}

//	Returns the button status of the specified joystick
static word INL_GetJoyButtons(word joy)
{
/*
register	word	result;

	result = inportb(0x201);	// Get all the joystick buttons
	result >>= joy? 6 : 4;	// Shift into bits 0-1
	result &= 3;				// Mask off the useless bits
	result ^= 3;
	return(result);
*/
return 0;
}

//	Detects and sets up the mouse
static wbool INL_StartMouse()
{
	return S_MouseExists();
}

//	Sets up scaling values for the specified joystick
static void INL_SetJoyScale(word joy)
{
	JoystickDef	*def;

	def = &JoyDefs[joy];
	def->joyMultXL = JoyScaleMax / (def->threshMinX - def->joyMinX);
	def->joyMultXH = JoyScaleMax / (def->joyMaxX - def->threshMaxX);
	def->joyMultYL = JoyScaleMax / (def->threshMinY - def->joyMinY);
	def->joyMultYH = JoyScaleMax / (def->joyMaxY - def->threshMaxY);
}

//	Sets up thresholding values and calls INL_SetJoyScale()
//	to set up scaling values
void IN_SetupJoy(word joy,word minx,word maxx,word miny,word maxy)
{
	word d, r;
	JoystickDef *def;

	def = &JoyDefs[joy];

	def->joyMinX = minx;
	def->joyMaxX = maxx;
	r = maxx - minx;
	d = r / 3;
	def->threshMinX = ((r / 2) - d) + minx;
	def->threshMaxX = ((r / 2) + d) + minx;

	def->joyMinY = miny;
	def->joyMaxY = maxy;
	r = maxy - miny;
	d = r / 3;
	def->threshMinY = ((r / 2) - d) + miny;
	def->threshMaxY = ((r / 2) + d) + miny;

	INL_SetJoyScale(joy);
}

//	Detects & auto-configures the specified joystick
//	The auto-config assumes the joystick is centered
static wbool INL_StartJoy(word joy)
{
	word x, y;

	if (S_JoystickExists(joy))
	{
		IN_GetJoyAbs(joy, &x, &y);
		IN_SetupJoy(joy, 0, 1000*2, 0, 1000*2);

		return true;
	}

	return false;
}

//	Starts up the Input Mgr
void IN_Startup()
{
	wbool checkjoys, checkmouse;
	word i;
	extern int _argc;
	extern char** _argv;

	if (IN_Started)
		return;

	checkjoys = true;
	checkmouse = true;
	for (i = 1; i < _argc; i++)
	{
		switch (US_CheckParm(_argv[i], ParmStrings))
		{
		case 0:
			checkjoys = false;
			break;
		case 1:
			checkmouse = false;
		break;
		}
	}

	IN_ClearKeysDown();

	MousePresent = checkmouse ? INL_StartMouse() : false;

	for (i = 0; i < MaxJoys; i++)
		JoysPresent[i] = checkjoys ? INL_StartJoy(i) : false;

	IN_Started = true;
}

//	Sets up default conditions for the Input Mgr
void IN_Default(wbool gotit, ControlType in)
{
	if ((!gotit)
	|| 	((in == ctrl_Joystick1) && !JoysPresent[0])
	|| 	((in == ctrl_Joystick2) && !JoysPresent[1])
	|| 	((in == ctrl_Mouse) && !MousePresent))
		in = ctrl_Keyboard1;

	Controls[0] = in;
}

//	Shuts down the Input Mgr
void IN_Shutdown()
{
	word i;

	if (!IN_Started)
		return;

	for (i = 0; i < MaxJoys; i++)
		JoysPresent[i] = false;

	IN_Started = false;
}

//	Clears the keyboard array
void IN_ClearKeysDown()
{
	LastScan = sc_None;
	LastASCII = key_None;
	memset(Keyboard,0,sizeof(Keyboard));
}

//	Reads the device associated with the specified
//	player and fills in the control info struct
void IN_ReadControl(int player, ControlInfo *info)
{
	wbool		realdelta = false;
	byte		dbyte;
	word		buttons;
	int			dx,dy;
	Motion		mx,my;
	ControlType	type;
	register	KeyboardDef	*def;

	dx = dy = 0;
	mx = my = motion_None;
	buttons = 0;
	S_ProcessEvents();

	if (DemoMode == demo_Playback)
	{
		dbyte = DemoBuffer[DemoOffset + 1];
		my = (dbyte & 3) - 1;
		mx = ((dbyte >> 2) & 3) - 1;
		buttons = (dbyte >> 4) & 3;

		if (!(--DemoBuffer[DemoOffset]))
		{
			DemoOffset += 2;
			if (DemoOffset >= DemoSize)
				DemoMode = demo_PlayDone;
		}

		realdelta = false;
	}
	else if (DemoMode == demo_PlayDone)
		Quit("Demo playback exceeded");
	else
	{
		switch (type = Controls[player])
		{
		case ctrl_Keyboard:
			def = &KbdDefs;

			if (Keyboard[def->upleft])
				mx = motion_Left,my = motion_Up;
			else if (Keyboard[def->upright])
				mx = motion_Right,my = motion_Up;
			else if (Keyboard[def->downleft])
				mx = motion_Left,my = motion_Down;
			else if (Keyboard[def->downright])
				mx = motion_Right,my = motion_Down;

			if (Keyboard[def->up])
				my = motion_Up;
			else if (Keyboard[def->down])
				my = motion_Down;

			if (Keyboard[def->left])
				mx = motion_Left;
			else if (Keyboard[def->right])
				mx = motion_Right;

			if (Keyboard[def->button0])
				buttons += 1 << 0;
			if (Keyboard[def->button1])
				buttons += 1 << 1;
			realdelta = false;
			break;
		case ctrl_Joystick1:
		case ctrl_Joystick2:
			INL_GetJoyDelta(type - ctrl_Joystick,&dx,&dy);
			buttons = INL_GetJoyButtons(type - ctrl_Joystick);
			realdelta = true;
			break;
		case ctrl_Mouse:
			INL_GetMouseDelta(&dx,&dy);
			buttons = INL_GetMouseButtons();
			realdelta = true;
			break;
		case ctrl_Keyboard2:
			break;
		}
	}

	if (realdelta)
	{
		mx = (dx < 0)? motion_Left : ((dx > 0)? motion_Right : motion_None);
		my = (dy < 0)? motion_Up : ((dy > 0)? motion_Down : motion_None);
	}
	else
	{
		dx = mx * 127;
		dy = my * 127;
	}

	info->x = dx;
	info->xaxis = mx;
	info->y = dy;
	info->yaxis = my;
	info->button0 = buttons & (1 << 0);
	info->button1 = buttons & (1 << 1);
	info->button2 = buttons & (1 << 2);
	info->button3 = buttons & (1 << 3);
	info->dir = DirTable[((my + 1) * 3) + (mx + 1)];

	if (DemoMode == demo_Record)
	{
		// Pack the control info into a byte
		dbyte = (buttons << 4) | ((mx + 1) << 2) | (my + 1);

		if ((DemoBuffer[DemoOffset + 1] == dbyte) &&
			(DemoBuffer[DemoOffset] < 255))
			(DemoBuffer[DemoOffset])++;
		else
		{
			if (DemoOffset || DemoBuffer[DemoOffset])
				DemoOffset += 2;

			if (DemoOffset >= DemoSize)
				Quit("Demo buffer overflow");

			DemoBuffer[DemoOffset] = 1;
			DemoBuffer[DemoOffset + 1] = dbyte;
		}
	}
}

//	Waits for an ASCII char, then clears LastASCII and
//	returns the ASCII value
char IN_WaitForASCII()
{
	char result;

	while (!(result = LastASCII))
		S_ProcessEvents();

	LastASCII = '\0';
	return(result);
}

//	waits for a button or key press.  If a button is down, upon
//	calling, it must be released for it to be recognized

wbool	btnstate[8];

void IN_StartAck()
{
	unsigned short i, buttons;

	// get initial state of everything
	IN_ClearKeysDown();
	memset(btnstate,0,sizeof(btnstate));

	buttons = IN_JoyButtons() << 4;
	if (MousePresent)
		buttons |= IN_MouseButtons();

	for (i = 0; i < 8; i++, buttons>>=1)
		if (buttons&1)
			btnstate[i] = true;
}

wbool IN_CheckAck()
{
	unsigned i, buttons;

	S_ProcessEvents();

	// see if something has been pressed
	if (LastScan)
		return true;

	buttons = IN_JoyButtons () << 4;
	if (MousePresent)
		buttons |= IN_MouseButtons ();

	for (i=0;i<8;i++,buttons>>=1)
		if (buttons & 1)
		{
			if (!btnstate[i])
				return true;
		}
		else
			btnstate[i]=false;

	return false;
}

void IN_Ack()
{
	IN_StartAck();

	while (!IN_CheckAck())
		;
}

//	Waits for the specified delay time (in ticks) or the
//	user pressing a key or a mouse button. If the clear flag is set, it
//	then either clears the key or waits for the user to let the mouse
//	button up.
wbool IN_UserInput(longword delay)
{
	longword elapsed;

	elapsed = 0;
	IN_StartAck ();
	do
	{
		if (IN_CheckAck())
			return true;
		elapsed += S_GetNewTics();
	} while (elapsed < delay);

	return false;
}

//===========================================================================

byte IN_MouseButtons()
{
	if (MousePresent)
		return S_PollMouseButtons();
	else
		return 0;
}

byte IN_JoyButtons()
{
	return S_PollJoyButtons();
}
