//
// Simplified Memory Manager, just wrap malloc and free
//

#include <stdlib.h>
#include <string.h>
#include "id_heads.h"

//=============================================================================
//							LOCAL INFO
//=============================================================================

#define LOCKBIT		0x80	// if set in attributes, block cannot be moved
#define PURGEBITS	3		// 0-3 level, 0= unpurgable, 3= purge first
#define PURGEMASK	0xfffc
#define BASEATTRIBUTES	0	// unlocked, non purgable

typedef struct mmblockstruct
{
  unsigned short attributes;
  memptr *useptr; // pointer to the segment start
  struct mmblockstruct *next;
} mmblocktype;

//=============================================================================
//						 GLOBAL VARIABLES
//=============================================================================

mminfotype	mminfo;
memptr		bufferseg;
wbool		mmerror;

void (* beforesort)(void);
void (* aftersort)(void);

//=============================================================================
//						 LOCAL VARIABLES
//=============================================================================

wbool		mmstarted;
wbool		bombonerror;
void		*nearheap;

mmblocktype	*mmblocks;

//==========================================================================

// Allocates bufferseg misc buffer
void MM_Startup()
{
  if (mmstarted)
    MM_Shutdown();

  mmstarted = true;
  bombonerror = true;

  // set up the linked list
  mmblocks = NULL;

  // Fill with any value
  mminfo.nearheap = 4096*1024;
  mminfo.mainmem = mminfo.nearheap;

  MM_GetPtr (&bufferseg,BUFFERSIZE);
}

//==========================================================================

// Frees all memory allocated
void MM_Shutdown()
{
  if (!mmstarted)
    return;

  while (mmblocks)
    MM_FreePtr (mmblocks->useptr);
}

//==========================================================================

//	Allocates an unlocked, unpurgable block
void MM_GetPtr (memptr *baseptr, unsigned long size)
{
  mmblocktype *mmnew;

  mmnew = malloc (sizeof (mmblocktype));

  mmnew->next = mmblocks;
  mmblocks = mmnew;

  mmnew->useptr = baseptr;
  mmnew->attributes = BASEATTRIBUTES;

  *baseptr = malloc (size);

  if (*baseptr == NULL)
    Quit ("MM_GetPtr: Out of memory!");
}

//==========================================================================

// Deallocates an unlocked, purgable block
void MM_FreePtr(memptr *baseptr)
{
  mmblocktype *scan = mmblocks, *last = NULL;

  while (scan != NULL)
  {
    if (*scan->useptr == *baseptr)
    {
      free (*baseptr);
      *baseptr = NULL;
      if (last)
	last->next = scan->next;
      else
	mmblocks = scan->next;
      free (scan);
      return;
    }

    last = scan;
    scan = scan->next;
  }

  Quit ("MM_FreePtr: Block not found!");
}

//==========================================================================

// Sets the purge level for a block (locked blocks cannot be made purgable)
void MM_SetPurge (memptr *baseptr, int purge)
{
  mmblocktype *scan;

  for (scan = mmblocks; scan != NULL; scan = scan->next)
  {
    if (*scan->useptr == *baseptr)
    {
      scan->attributes &= ~PURGEBITS;
      scan->attributes |= purge;
      return;
    }
  }

  Quit ("MM_SetPurge: Block not found!");
}

//==========================================================================

// Locks / unlocks the block
void MM_SetLock (memptr *baseptr, wbool locked)
{
  mmblocktype *scan;

  for (scan = mmblocks; scan != NULL; scan = scan->next)
  {
    if (scan->useptr == baseptr)
    {
      scan->attributes &= ~LOCKBIT;
      scan->attributes |= locked*LOCKBIT;
      return;
    }
  }

  Quit ("MM_SetLock: Block not found!");
}

//==========================================================================

// Throws out all purgable stuff and compresses movable blocks
void MM_SortMem()
{
  mmblocktype *scan, *last, *next;
  unsigned long start = 0,length,source,dest;
  int playing;

  // lock down a currently playing sound
  playing = SD_SoundPlaying();
  if (playing)
  {
    switch (SoundMode)
    {
    case sdm_Off:
      break;
    case sdm_PC:
      playing += STARTPCSOUNDS;
      break;
    case sdm_AdLib:
      playing += STARTADLIBSOUNDS;
      break;
    }
    MM_SetLock((memptr)&audiosegs[playing],true);
  }

  SD_StopSound();

  if (beforesort)
    beforesort();

  scan = mmblocks;
  last = NULL;

  while (scan)
  {
    if (scan->attributes & PURGEBITS)
    {
      // throw out the purgable block
      next = scan->next;
      free (*scan->useptr);
      *scan->useptr = NULL;
      free (scan);
      if (last)
	last->next = next;
      else
	mmblocks = next;
      scan = next;
      continue;
    }

    last = scan;
    scan = scan->next;	// go to next block
  }

  if (aftersort)
    aftersort();

  if (playing)
    MM_SetLock((memptr)&audiosegs[playing],false);
}

//==========================================================================

// Returns the total free space without purging
long MM_UnusedMemory()
{
  return 4096*1024;
}

// Returns the total free space with purging
long MM_TotalFree()
{
  return 4096*1024;
}

void MM_BombOnError(wbool bomb)
{
  bombonerror = bomb;
}

